﻿--[[--------------------------------------------------------------------------
  Copyright (c) 2009, Nathan Pieper
  All rights reserved.

  ShadowSpell is an interactive chat text spell checker for World of Warcraft
--------------------------------------------------------------------------]]--

ShadowSpell = {
	VERSION = "ShadowSpell 1.0 Interactive Spellcheck",
	queue = {},
	queuePos = 0,
}
local L = ShadowSpellLocals

ShadowSpell_DB = {

}

local DB = {}
local BFTable = {}
local BFBits = 2251876
local BFNumHashes = 7



function ShadowSpell:split(str, patt)
	vals = {}; valindex = 0; word = ""
	-- need to add a trailing separator to catch the last value.
	str = str .. patt
	for i = 1, string.len(str) do

		cha = string.sub(str, i, i)
		if cha ~= patt then
			word = word .. cha
		else
			if word ~= nil then
				vals[valindex] = word
				valindex = valindex + 1
				word = ""
			else
				-- in case we get a line with no data.
				break
			end
		end

	end
	return vals
end


function bestK(numElements, errorRate)
	val0 = math.log(2.0) * bestM(numElements, errorRate) / numElements
	val1, r = math.modf(val0)
	if r >= .5 then
		val1 = val1 + 1
	end

	--val1, val2 = math.modf(math.round(math.log(2.0) * bestM(numElements, errorRate) / numElements))
	return val1
end

function bestM(numElements, errorRate)
	optimalRateBase = 1.0 / math.pow(2, math.log(2.0))
	val1, val2 = math.modf(math.ceil(numElements * math.log(errorRate)/math.log(optimalRateBase)))
	return val1
end

function GetBit(bitOrdinal, data)
	return bit.band(bit.rshift(data, bitOrdinal),1) == 1
end

--Set a bit in an integer to 1
function SetBit(bitOrdinal, data)
	return bit.bor(data, bit.lshift(1, bitOrdinal))
end

function InitBFTable(numElements, errorRate)
	bitsNeeded = bestM(numElements, errorRate)
	BFBits = bitsNeeded

	--Store the bits in and array of 32 bit integers
	tableSize = math.ceil(bitsNeeded / 32)

	BFNumHashes = bestK(numElements, errorRate)

	--Build the table
	for x = 0, tableSize do
		BFTable[x] = nil
	end

	--Make sure the last element isn't nil so we can serialize the entire array
	BFTable[tableSize+1] = 0

	print("bitsNeeded:", bitsNeeded)
	print("tableSize:", tableSize)
end

--<summary>
--Hashes a string using Bob Jenkin's "One At A Time" method from Dr. Dobbs (http://burtleburtle.net/bob/hash/doobs.html).
--Runtime is suggested to be 9x+9, where x = input.Length.
--</summary>
--<param name="input">The string to hash.</param>
--<returns>The hashed result.</returns>
function hashString2(input)
	s = input
	hash = 0

	for i = 0, #input - 1 do
		hash = hash + string.byte(i+1)
		hash = hash + bit.lshift(hash, 10)
		hash = math.pow(hash, bit.rshift(hash, 6))
	end

	hash = hash + bit.lshift(hash, 3)
	hash = math.pow(hash, bit.rshift(hash, 11))
	hash = hash + bit.lshift(hash, 15)

	return hash
end

--[[/*
---------------------------------------------------------------------------
Copyright (c) 2003, Dominik Reichl <dominik.reichl@t-online.de>, Germany.
All rights reserved.

Distributed under the terms of the GNU General Public License v2.

This software is provided 'as is' with no explicit or implied warranties
in respect of its properties, including, but not limited to, correctness
and/or fitness for purpose.
---------------------------------------------------------------------------
*/]]

--// FCS-32 algorithm implemented as described in RFC 1331
--// Small changes made to remove init routine

local FCSINIT32 = -1

--// Fast 32 bit FCS lookup table
local fcs32tab = {  [0]=0, 1996959894, -301047508, -1727442502, 124634137, 1886057615, -379345611, -1637575261,
	249268274, 2044508324, -522852066, -1747789432, 162941995, 2125561021, -407360249, -1866523247,
	498536548, 1789927666, -205950648, -2067906082, 450548861, 1843258603, -187386543, -2083289657,
	325883990, 1684777152, -43845254, -1973040660, 335633487, 1661365465, -99664541, -1928851979,
	997073096, 1281953886, -715111964, -1570279054, 1006888145, 1258607687, -770865667, -1526024853,
	901097722, 1119000684, -608450090, -1396901568, 853044451, 1172266101, -589951537, -1412350631,
	651767980, 1373503546, -925412992, -1076862698, 565507253, 1454621731, -809855591, -1195530993,
	671266974, 1594198024, -972236366, -1324619484, 795835527, 1483230225, -1050600021, -1234817731,
	1994146192, 31158534, -1731059524, -271249366, 1907459465, 112637215, -1614814043, -390540237,
	2013776290, 251722036, -1777751922, -519137256, 2137656763, 141376813, -1855689577, -429695999,
	1802195444, 476864866, -2056965928, -228458418, 1812370925, 453092731, -2113342271, -183516073,
	1706088902, 314042704, -1950435094, -54949764, 1658658271, 366619977, -1932296973, -69972891,
	1303535960, 984961486, -1547960204, -725929758, 1256170817, 1037604311, -1529756563, -740887301,
	1131014506, 879679996, -1385723834, -631195440, 1141124467, 855842277, -1442165665, -586318647,
	1342533948, 654459306, -1106571248, -921952122, 1466479909, 544179635, -1184443383, -832445281,
	1591671054, 702138776, -1328506846, -942167884, 1504918807, 783551873, -1212326853, -1061524307,
	-306674912, -1698712650, 62317068, 1957810842, -355121351, -1647151185, 81470997, 1943803523,
	-480048366, -1805370492, 225274430, 2053790376, -468791541, -1828061283, 167816743, 2097651377,
	-267414716, -2029476910, 503444072, 1762050814, -144550051, -2140837941, 426522225, 1852507879,
	-19653770, -1982649376, 282753626, 1742555852, -105259153, -1900089351, 397917763, 1622183637,
	-690576408, -1580100738, 953729732, 1340076626, -776247311, -1497606297, 1068828381, 1219638859,
	-670225446, -1358292148, 906185462, 1090812512, -547295293, -1469587627, 829329135, 1181335161,
	-882789492, -1134132454, 628085408, 1382605366, -871598187, -1156888829, 570562233, 1426400815,
	-977650754, -1296233688, 733239954, 1555261956, -1026031705, -1244606671, 752459403, 1541320221,
	-1687895376, -328994266, 1969922972, 40735498, -1677130071, -351390145, 1913087877, 83908371,
	-1782625662, -491226604, 2075208622, 213261112, -1831694693, -438977011, 2094854071, 198958881,
	-2032938284, -237706686, 1759359992, 534414190, -2118248755, -155638181, 1873836001, 414664567,
	-2012718362, -15766928, 1711684554, 285281116, -1889165569, -127750551, 1634467795, 376229701,
	-1609899400, -686959890, 1308918612, 956543938, -1486412191, -799009033, 1231636301, 1047427035,
	-1362007478, -640263460, 1088359270, 936918000, -1447252397, -558129467, 1202900863, 817233897,
	-1111625188, -893730166, 1404277552, 615818150, -1160759803, -841546093, 1423857449, 601450431,
	-1285129682, -1000256840, 1567103746, 711928724, -1274298825, -1022587231, 1510334235, 755167117 }


function hashString1(pBuffer)
	uFcs32 = FCSINIT32
	local i
	local l=string.len(pBuffer)
	for i = 1,l do
		uFcs32 = bit.bxor(bit.rshift(uFcs32,8), fcs32tab[bit.band(bit.bxor(uFcs32, string.byte(pBuffer,i)), 255)])
	end
	uFcs32 = bit.bnot(uFcs32)
	return uFcs32
end


--<summary>
--Performs Dillinger and Manolios double hashing.
--</summary>
function computeHash(primaryHash, secondaryHash, i)  --As Integer
	resultingHash = (primaryHash + (i * secondaryHash)) % BFBits
    return math.abs(resultingHash)
end

function Insert(input)
	h1 = hashString1(string.lower(input))
	h2 = hashString2(string.lower(input))

	for i = 0, BFNumHashes do
		hash = computeHash(h1, h2, i)

		--print("Hash:", hash)

		--Row
		row = math.modf(hash / 32)
		--Bit
		bitToSet = hash % 32

		--print("i:", i, "Row:", row, "Insert Row Prev:", BFTable[row])

		if BFTable[row] == nil then
			BFTable[row] = 0
		end

		--Store 32bit numbers as hex to save space.
		h = tonumber("0x"..tostring(BFTable[row]))
		h = SetBit(bitToSet, h)

		BFTable[row] = string.format("%x", h)
		--BFTable[row] = SetBit(bitToSet, BFTable[row])

		--print("i:", i, "Row:", row, "Insert Row After:", BFTable[row])
	end
end

function Contains(input)
	if input == nui then
		return true
	end

	h1 = hashString1(string.lower(input))
	h2 = hashString2(string.lower(input))

	print("h1", h1, "h2", h2)

	for i = 0, BFNumHashes do
		hash = computeHash(h1, h2, i)
		print("ComputeHash #",i , hash)
		--Row
		row = math.modf(hash / 32)
		--Bit
		bitToTest = hash % 32

		if BFTable[row] == nil then
			return false
		end

		h = tonumber("0x"..tostring(BFTable[row]))
		if GetBit(bitToTest, h) == false then
			return false
		end
	end
	return true
end

--- **AceSerializer-3.0** can serialize any variable (except functions or userdata) into a string format,
-- that can be send over the addon comm channel. AceSerializer was designed to keep all data intact, especially
-- very large numbers or floating point numbers, and table structures. The only caveat currently is, that multiple
-- references to the same table will be send individually.
--
-- **AceSerializer-3.0** can be embeded into your addon, either explicitly by calling AceSerializer:Embed(MyAddon) or by
-- specifying it as an embeded library in your AceAddon. All functions will be available on your addon object
-- and can be accessed directly, without having to explicitly call AceSerializer itself.\\
-- It is recommended to embed AceSerializer, otherwise you'll have to specify a custom `self` on all calls you
-- make into AceSerializer.
-- @class file
-- @name AceSerializer-3.0
-- @release $Id: AceSerializer-3.0.lua 768 2009-04-04 10:18:38Z nevcairiel $
local MAJOR,MINOR = "AceSerializer-3.0", 2
local AceSerializer = {}

--if not AceSerializer then return end

local strbyte = string.byte
local strchar = string.char
local tconcat = table.concat
local gsub = string.gsub
local gmatch = string.gmatch
local pcall = pcall
local format = string.format
local type = type
local tostring, tonumber = tostring, tonumber
local select = select

-- quick copies of string representations of wonky numbers
local serNaN = tostring(0/0)
local serInf = tostring(1/0)
local serNegInf = tostring(-1/0)


-- Serialization functions

local function SerializeStringHelper(ch)	-- Used by SerializeValue for strings
	-- We use \126 ("~") as an escape character for all nonprints plus a few more
	local n = strbyte(ch)
	if n<=32 then 			-- nonprint + space
		return "\126"..strchar(n+64)
	elseif n==94 then		-- value separator
		return "\126\125"
	elseif n==126 then		-- our own escape character
		return "\126\124"
	elseif n==127 then		-- nonprint (DEL)
		return "\126\123"
	else
		assert(false)	-- can't be reached if caller uses a sane regex
	end
end

local function SerializeValue(v, res, nres)
	-- We use "^" as a value separator, followed by one byte for type indicator
	local t=type(v)

	if t=="string" then		-- ^S = string (escaped to remove nonprints, "^"s, etc)
		res[nres+1] = "^S"
		res[nres+2] = gsub(v,"[%c \94\126\127]", SerializeStringHelper)
		nres=nres+2

	elseif t=="number" then	-- ^N = number (just tostring()ed) or ^F (float components)
		local str = tostring(v)
		if tonumber(str)==v  or str==serNaN or str==serInf or str==serNegInf then
			-- translates just fine, transmit as-is
			res[nres+1] = "^N"
			res[nres+2] = str
			nres=nres+2
		else
			local m,e = frexp(v)
			res[nres+1] = "^F"
			res[nres+2] = format("%.0f",m*2^53)	-- force mantissa to become integer (it's originally 0.5--0.9999)
			res[nres+3] = "^f"
			res[nres+4] = tostring(e-53)	-- adjust exponent to counteract mantissa manipulation
			nres=nres+4
		end

	elseif t=="table" then	-- ^T...^t = table (list of key,value pairs)
		nres=nres+1
		res[nres] = "^T"
		for k,v in pairs(v) do
			nres = SerializeValue(k, res, nres)
			nres = SerializeValue(v, res, nres)
		end
		nres=nres+1
		res[nres] = "^t"

	elseif t=="boolean" then	-- ^B = true, ^b = false
		nres=nres+1
		if v then
			res[nres] = "^B"	-- true
		else
			res[nres] = "^b"	-- false
		end

	elseif t=="nil" then		-- ^Z = nil (zero, "N" was taken :P)
		nres=nres+1
		res[nres] = "^Z"

	else
		error(MAJOR..": Cannot serialize a value of type '"..t.."'")	-- can't produce error on right level, this is wildly recursive
	end

	return nres
end



local serializeTbl = { "^1" }	-- "^1" = Hi, I'm data serialized by AceSerializer protocol rev 1

--- Serialize the data passed into the function.
-- Takes a list of values (strings, numbers, booleans, nils, tables)
-- and returns it in serialized form (a string).\\
-- May throw errors on invalid data types.
-- @param ... List of values to serialize
-- @return The data in its serialized form (string)
function AceSerializer:Serialize(...)
	local nres = 1

	for i=1,select("#", ...) do
		local v = select(i, ...)
		nres = SerializeValue(v, serializeTbl, nres)
	end

	serializeTbl[nres+1] = "^^"	-- "^^" = End of serialized data

	return tconcat(serializeTbl, "", 1, nres+1)
end

-- Deserialization functions
local function DeserializeStringHelper(escape)
	if escape<"~\123" then
		return strchar(strbyte(escape,2,2)-64)
	elseif escape=="~\123" then
		return "\127"
	elseif escape=="~\124" then
		return "\126"
	elseif escape=="~\125" then
		return "\94"
	end
	error("DeserializeStringHelper got called for '"..escape.."'?!?")  -- can't be reached unless regex is screwed up
end

local function DeserializeNumberHelper(number)
	if number == serNaN then
		return 0/0
	elseif number == serNegInf then
		return -1/0
	elseif number == serInf then
		return 1/0
	else
		return tonumber(number)
	end
end

-- DeserializeValue: worker function for :Deserialize()
-- It works in two modes:
--   Main (top-level) mode: Deserialize a list of values and return them all
--   Recursive (table) mode: Deserialize only a single value (_may_ of course be another table with lots of subvalues in it)
--
-- The function _always_ works recursively due to having to build a list of values to return
--
-- Callers are expected to pcall(DeserializeValue) to trap errors

local function DeserializeValue(iter,single,ctl,data)

	if not single then
		ctl,data = iter()
	end

	if not ctl then
		error("Supplied data misses AceSerializer terminator ('^^')")
	end

	if ctl=="^^" then
		-- ignore extraneous data
		return
	end

	local res

	if ctl=="^S" then
		res = gsub(data, "~.", DeserializeStringHelper)
	elseif ctl=="^N" then
		res = DeserializeNumberHelper(data)
		if not res then
			error("Invalid serialized number: '"..tostring(data).."'")
		end
	elseif ctl=="^F" then     -- ^F<mantissa>^f<exponent>
		local ctl2,e = iter()
		if ctl2~="^f" then
			error("Invalid serialized floating-point number, expected '^f', not '"..tostring(ctl2).."'")
		end
		local m=tonumber(data)
		e=tonumber(e)
		if not (m and e) then
			error("Invalid serialized floating-point number, expected mantissa and exponent, got '"..tostring(m).."' and '"..tostring(e).."'")
		end
		res = m*(2^e)
	elseif ctl=="^B" then	-- yeah yeah ignore data portion
		res = true
	elseif ctl=="^b" then   -- yeah yeah ignore data portion
		res = false
	elseif ctl=="^Z" then	-- yeah yeah ignore data portion
		res = nil
	elseif ctl=="^T" then
		-- ignore ^T's data, future extensibility?
		res = {}
		local k,v
		while true do
			ctl,data = iter()
			if ctl=="^t" then break end	-- ignore ^t's data
			k = DeserializeValue(iter,true,ctl,data)
			if k==nil then
				error("Invalid AceSerializer table format (no table end marker)")
			end
			ctl,data = iter()
			v = DeserializeValue(iter,true,ctl,data)
			if v==nil then
				error("Invalid AceSerializer table format (no table end marker)")
			end
			res[k]=v
		end
	else
		error("Invalid AceSerializer control code '"..ctl.."'")
	end

	if not single then
		return res,DeserializeValue(iter)
	else
		return res
	end
end

--- Deserializes the data into its original values.
-- Accepts serialized data, ignoring all control characters and whitespace.
-- @param str The serialized data (from :Serialize)
-- @return true followed by a list of values, OR false followed by an error message
function AceSerializer:Deserialize(str)
	str = gsub(str, "[%c ]", "")	-- ignore all control characters; nice for embedding in email and stuff

	local iter = gmatch(str, "(^.)([^^]*)")	-- Any ^x followed by string of non-^
	local ctl,data = iter()
	if not ctl or ctl~="^1" then
		-- we purposefully ignore the data portion of the start code, it can be used as an extension mechanism
		return false, "Supplied data is not AceSerializer data (rev 1)"
	end

	return pcall(DeserializeValue, iter)
end


----------------------------------------
-- Base library stuff
----------------------------------------

AceSerializer.internals = {	-- for test scripts
	SerializeValue = SerializeValue,
	SerializeStringHelper = SerializeStringHelper,
}

local mixins = {
	"Serialize",
	"Deserialize",
}

AceSerializer.embeds = AceSerializer.embeds or {}

function AceSerializer:Embed(target)
	for k, v in pairs(mixins) do
		target[v] = self[v]
	end
	self.embeds[target] = true
	return target
end

-- Update embeds
for target, v in pairs(AceSerializer.embeds) do
	AceSerializer:Embed(target)
end









function ShadowSpell:Initialize(frame)
	ShadowSpell:OnSizeChanged(frame)
	table.insert(UISpecialFrames, "ShadowSpellFrame")
	PlaySound("igMainMenuOpen");
	self:UpdateButtons()
end

function ShadowSpell:Button_OnEnter(frame)

end

function ShadowSpell:Button_OnLeave(frame)
	GameTooltip:Hide()
end

function ShadowSpell:Button_OnClick(button)
	local operation = button:GetName():match("ShadowSpellButton(.+)")
	if operation == "New" then
		ShadowSpell:Button_New(button)
	elseif operation == "Open" then
		ShadowSpell:Button_Open(button)
	elseif operation == "Save" then
		ShadowSpell:Button_Save(button)
	elseif operation == "Undo" then
		ShadowSpell:Button_Undo(button)
	elseif operation == "Redo" then
		ShadowSpell:Button_Redo(button)
	elseif operation == "Delete" then
		ShadowSpell:Button_Delete(button)
	elseif operation == "Lock" then
		ShadowSpell:Button_Lock(button)
	elseif operation == "Unlock" then
		ShadowSpell:Button_Unlock(button)
	elseif operation == "Previous" then
		ShadowSpell:Button_Previous(button)
	elseif operation == "Next" then
		ShadowSpell:Button_Next(button)
	elseif operation == "Run" then
		ShadowSpell:Button_Run(button)
	elseif operation == "Close" then
		ShadowSpell:Button_Close(button)
	end
end

function ShadowSpell_OnLoad()
	this:RegisterEvent("VARIABLES_LOADED");
	this:RegisterEvent("ADDON_LOADED");

end

function ShadowSpell_OnEvent()
	if ( event == "VARIABLES_LOADED" ) then
		ShadowSpell_VariablesLoaded();
	elseif( event == "ADDON_LOADED" ) then
		ShadowSpell_CompatibilityCheck();
	end
end

function ShadowSpell_VariablesLoaded()
	ret, BFTable = AceSerializer:Deserialize(BFDict)
	print("SS: Vars Loaded", #BFTable)
	hooksecurefunc("ChatEdit_OnSpacePressed", ShadowSpell_OnSpacePressed)
end

function ShadowSpell_CompatibilityCheck()
	--print("SS: Addon_Loaded")
end

function ShadowSpell_OnSpacePressed()
	local text = ChatFrameEditBox:GetText()
	print("Space pressed:", text)
	words = ShadowSpell:split(text, " ")
	len = #words
	print("Checking word:", words[len-1], "Correct: ", Contains(words[len-1]))
end

function ShadowSpell:OnSizeChanged(frame)

end

